#!/usr/bin/env python
# -*- coding:UTF-8 -*-

import argparse
import re

from .math_utils import *


def parse_dict(s):
    try:
        d = eval(s)
        if not isinstance(d, dict):
            raise argparse.ArgumentTypeError(f"Invalid dictionary format: '{s}'")
        return d
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid dictionary format: '{s}'")

def safe_compile(code):
    """
    The eval function that filters out harmful strings

    Args:
        code ():

    Returns:
        code: a complied string
    """
    banned = ('eval', 'compile', 'exec', 'getattr', 'hasattr', 'setattr', 'delattr',
              'classmethod', 'globals', 'help', 'input', 'isinstance', 'issubclass', 'locals',
              'open', 'print', 'property', 'staticmethod', 'vars')

    # Check for banned names in the code
    for name in re.findall(r'\b\w+\b', code):
        if name in banned:
            raise NameError(f'{name} not allowed : arbitrary code execution not allowed')

    # Compile and evaluate the code
    return compile(code, '', 'eval')

def eval_with_vars(code, geo, constant, outs, dic):
    """
    Run a code string with dictionaries as vars.

    Args:
        code (str): code to be executed
        geo (dict[str, Tensor]): dict with input features' names and their values
        constant (dict[str, int | float]): dict with constants' names and their values
        outs (dict[str, Tensor]): dict with output features' names and their values
        dic (dict[str, Tensor]): dict with gradients or intermediate variables' names and their values

    Returns:
        Any: result of the code string
    """
    return eval(code)

def get_grad(variable: str, geo, dic, outs):
    """
    Get grad through backend's automatic differentiation.

    Args:
        variable (str): code to be executed
        geo (dict[str, Tensor]): dict with input features' names and their values
        dic (dict[str, int | float]): dict with calculated gradients' names and their values
        outs (dict[str, Tensor]): dict with output features' names and their values

    Returns:
        Any: result of the code string
    """
    y, _, x = variable.rpartition('_')

    return bkd.grad(outs[y], geo[x]) if '_' not in y else bkd.grad(dic[y], geo[x])

def get_grad2d(variable: str, geo, dic, outs, imgfilter=None):
    y, x = variable.split('_')
    if ".h" in variable:
        return imgfilter.grad_h(outs[y])
    elif ".v" in variable:
        return imgfilter.grad_v(outs[y])
    else:
        pass

def convert_math_to_python(equation: str):
    def repl(match):
        return match.group(1) + ("_" + match.group(1)) * (int(match.group(2)) - 1)

    equation = equation.replace('[', '(')
    equation = equation.replace(']', ')')
    equation = equation.replace('^', '**')

    pattern = r"\s*{(.*?),\s*(\d+)}"
    while re.search(pattern, equation):
        equation = re.sub(pattern, repl, equation)

    pattern = r"D\((.*?),\s*(.*?)\)"
    while re.search(pattern, equation):
        equation = re.sub(pattern, r"\1_\2", equation)

    return equation

def equation_to_runnable(equation: str, geo, constant, outs, dic):
    """
    Replace the variable name into exist value in dicts

    Args:
        equation (str): a string contains user defined variable
        geo (dict[str, Tensor]): dict with input features' names and their values
        constant (dict[str, int | float]): dict with constants' names and their values
        outs (dict[str, Tensor]): dict with output features' names and their values
        dic (dict[str, Tensor]): dict with gradients or intermediate variables' names and their values
    Returns:
        str: an executable code string
    """
    equation = convert_math_to_python(equation)
    equation = re.sub(r"\s+", "", equation)
    variables = re.findall(r'\b([a-zA-Z_][a-zA-Z0-9_]*)\b', equation)

    original_pattern = lambda variable : r'\b' + variable + r'\b'
    replacement_pattern = {
        "geo" : lambda variable : "geo['%s']" % variable,
        "constant" : lambda variable : "constant['%s']" % variable,
        "outs" : lambda variable : "outs['%s']" % variable,
        "dic" : lambda variable : "dic['%s']" % variable,
    }
    replace = lambda variable: re.sub(original_pattern(variable), locate_var_in_dicts(variable), equation)
    # replace = lambda variable : equation.replace(original_pattern(variable), locate_var_in_dicts(variable))

    def locate_var_in_dicts(variable: str):
        if variable in geo:
            return replacement_pattern['geo'](variable)
        if variable in constant:
            return replacement_pattern['constant'](variable)
        if variable in outs:
            return replacement_pattern['outs'](variable)
        if '_' in variable:
            parts = variable.split('_')
            y_x = parts[0] + '_' + parts[1]
            dic[y_x] = None
            for x in parts[2:]:
                y_x += '_' + x
                dic[y_x] = None
        elif '|' in variable:
            dic[variable] = None
        return replacement_pattern['dic'](variable)

    for variable in variables:
        if variable not in math_list:
            equation = replace(variable)
    return equation

def split_pde(equation: str, geo, constant, outs, dic):
    code = equation_to_runnable(equation, geo, constant, outs, dic)
    left, right = code.split('=')
    return [safe_compile(left), safe_compile(right)]

def split_condition(equation: str, geo, constant, outs, dic):
    code = equation_to_runnable(equation, geo, constant, outs, dic)
    conds, whens = code.split('|')
    cond_list = []
    when_list = []
    for cond in conds.split(','):
        left, right = cond.split('=')
        cond_list.append([safe_compile(left), safe_compile(right)])

    for when in whens.split(','):
        when_left, when_right = when.split('=')
        when_list.append([when_left, safe_compile(when_right)])
    return [cond_list, when_list]

def split_hard_condition(equation: str, geo, constant, outs, dic):
    code = equation_to_runnable(equation, geo, constant, outs, dic)
    left, right = code.split('=')
    return [left, safe_compile(right)]

def loss_from_list(loss_func, data, weight):
    if weight:
        return sum((loss_func(pair[0], pair[1]) * weight).mean() for pair in data)
    return sum(loss_func(pair[0], pair[1]) for pair in data)

def loss_from_dict(loss_func, predict_dict, ground_truth_dict, weight):
    if weight:
        return sum((loss_func(value, predict_dict[key]) * weight).mean() for key, value in ground_truth_dict.items())
    return sum(loss_func(value, predict_dict[key]) for key, value in ground_truth_dict.items())